/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"

#include "apr_strings.h"
#include "apr_network_io.h"

#include "mod_smtpd.h"

module AP_MODULE_DECLARE_DATA smtpd_clamd_module;

typedef struct {
    int enable;

    char *host;
    int   port;

    int noreject;
} smtpd_clamd_config;

static void *smtpd_clamd_config_new(apr_pool_t *p,
				    server_rec *s)
{
    smtpd_clamd_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_clamd_config));

    cfg->enable = -1;
    cfg->port = -1;
    cfg->host = NULL;
    cfg->noreject = -1;

    return cfg;
}

static void *smtpd_clamd_config_merge(apr_pool_t *p, void *b,
				      void *n)
{
    smtpd_clamd_config *base = (smtpd_clamd_config *)b;
    smtpd_clamd_config *new = (smtpd_clamd_config *)n;

    smtpd_clamd_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_clamd_config));

    if ( new->enable == -1 ) {
	cfg->enable = base->enable;
    }
    else {
	cfg->enable = new->enable;
    }

    if ( new->noreject == -1 ) {
	cfg->noreject = base->noreject;
    }
    else {
	cfg->noreject = new->noreject;
    }

    if ( new->port == -1 ) {
	cfg->port = base->port;
    }
    else {
	cfg->port = new->port;
    }

    if ( new->host == NULL ) {
	if ( base->port != NULL ) {
	    cfg->host = apr_pstrdup(p, base->host);
	}
	else {
	    cfg->host = NULL;
	}
    }
    else {
	cfg->host = apr_pstrdup(p, new->host);
    }

    return cfg;
}

static const char *smtpd_clamd_enable(cmd_parms *cmd,
				      void *din_dconf, int flag)
{
    smtpd_clamd_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_clamd_module);

    conf->enable = flag;

    return NULL;
}

static const char *smtpd_clamd_host(cmd_parms *cmd,
				    void *din_dconf, const char *host)
{
    smtpd_clamd_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_clamd_module);

    conf->host = apr_pstrdup(cmd->pool, host);

    return NULL;
}

static const char *smtpd_clamd_port(cmd_parms *cmd,
				    void *din_dconf, const char *port)
{
    smtpd_clamd_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_clamd_module);

    conf->port = atoi(port);

    return NULL;
}

static const char *smtpd_clamd_noreject(cmd_parms *cmd,
					void *din_dconf, int flag)
{
    smtpd_clamd_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_clamd_module);

    conf->noreject = flag;

    return NULL;
}

static const command_rec smtpd_clamd_cmds[] =
{
    AP_INIT_FLAG("SmtpClamdEnable", smtpd_clamd_enable, (void *) NULL, RSRC_CONF,
		 "Enable clamav scanning via clamd"),
    AP_INIT_TAKE1("SmtpClamdHost", smtpd_clamd_host, (void *) NULL, RSRC_CONF,
		  "The host to connect to clamd on"),
    AP_INIT_TAKE1("SmtpClamdPort", smtpd_clamd_port, (void *) NULL, RSRC_CONF,
		  "The port to connect to clamd on"),
    AP_INIT_FLAG("SmtpClamdNoReject", smtpd_clamd_noreject, (void *) NULL, RSRC_CONF,
		 "Don't reject messages with virus"),
    {NULL}
};

smtpd_retcode smtpd_clamd(smtpd_conn_rec *scr, smtpd_return_data *in)
{
    smtpd_clamd_config *conf = ap_get_module_config(scr->s->module_config, &smtpd_clamd_module);
    smtpd_retcode ret = SMTPD_DECLINED;
    apr_pool_t *pool;
    apr_socket_t *sock_ctrl;
    apr_socket_t *sock_data;
    apr_sockaddr_t *sockaddr;

    int data_port;

    char buf[4096];
    apr_size_t buf_len;
    apr_size_t buf_size = sizeof(buf) - 1;
    char *str;

    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		 "smtpd_clamd");

    if (conf->enable == 0) {
	return SMTPD_DECLINED;
    }

    apr_pool_create(&pool, scr->p);

    if (apr_sockaddr_info_get(&sockaddr, conf->host, AF_INET, conf->port, 0, pool) != APR_SUCCESS)
    {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_clamd]: apr_sockaddr_info_get failed for %s:%d",
		     conf->host, conf->port);
	goto smtpd_clamd_cleanup;
    }

    if (apr_socket_create(&sock_ctrl, APR_INET, SOCK_STREAM, APR_PROTO_TCP, pool) != APR_SUCCESS)
    {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_clamd]: apr_socket_create failed for %s:%d",
		     conf->host, conf->port);
	goto smtpd_clamd_cleanup;
    }

    if (apr_socket_connect(sock_ctrl, sockaddr) != APR_SUCCESS)
    {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_clamd]: apr_socket_connect failed for %s:%d",
		     conf->host, conf->port);
	goto smtpd_clamd_cleanup;
    }

    buf_len = 6;
    apr_socket_send(sock_ctrl, "STREAM", &buf_len);

    buf_len = sizeof(buf)-1;
    apr_socket_recv(sock_ctrl, buf, &buf_len);

    buf[buf_len] = '\0';
    
    if ((str = strstr(buf, "PORT")) != NULL) {
	str += 5;
	sscanf(str, "%d", &data_port);
    }
    else {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_clamd]: error getting data port from clamd! %s",
		     buf);
	goto smtpd_clamd_cleanup_ctrl;
    }

    if (apr_sockaddr_info_get(&sockaddr, conf->host, AF_INET, data_port, 0, pool) != APR_SUCCESS)
    {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_clamd]: apr_sockaddr_info_get failed for data connection %s:%d",
		     conf->host, data_port);
	goto smtpd_clamd_cleanup_ctrl;
    }

    if (apr_socket_create(&sock_data, APR_INET, SOCK_STREAM, APR_PROTO_TCP, pool) != APR_SUCCESS)
    {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_clamd]: apr_socket_create failed for data connection %s:%d",
		     conf->host, data_port);
	goto smtpd_clamd_cleanup_ctrl;
    }

    if (apr_socket_connect(sock_data, sockaddr) != APR_SUCCESS)
    {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_clamd]: apr_socket_connect failed for data connection %s:%d",
		     conf->host, data_port);
	goto smtpd_clamd_cleanup_ctrl;
    }

    smtpd_body_resetpos(scr->transaction, SMTPD_HEADERS);

    while (smtpd_body_getline(scr->transaction, buf, sizeof(buf)) == APR_SUCCESS) {
	buf_len = strlen(buf);
	apr_socket_send(sock_data, buf, &buf_len);
    }

    apr_socket_close(sock_data);

    memset(buf, 0, sizeof(buf));
    buf_len = sizeof(buf)-1;
    str = buf;

    while (apr_socket_recv(sock_ctrl, str, &buf_len) == APR_SUCCESS && buf_len > 0 ) {
	str[buf_len] = '\0';
	str += buf_len;
	buf_len = buf_size - (str-buf);
    }

    if ((str = ap_strcasestr(buf, "FOUND\n")) != NULL) {
	char *virus;

	str--;
	*str = '\0';

	if ((virus = strchr(buf, ':')) != NULL) {
	    virus += 2;
	}
	else {
	    virus = buf;
	}

	if (conf->noreject != 1) {
	    char *msg = apr_psprintf(in->p, "virus %s detected by ClamAV - http://www.clamav.net", virus);
	    smtpd_respond_oneline(scr, 554, msg);
	    ret = SMTPD_DONE;

	    ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			 "[mod_smtpd_clamd]: virus %s detected", virus);
	}
	else {
	    ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			 "[mod_smtpd_clamd]: %s, but not rejecting the message", virus);
	}
    }
    else if (ap_strcasestr(buf, "ERROR\n") != NULL) {
	if (ap_strcasestr(buf, "Size limit reached") != NULL) {
	}

	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_clamd]: clamd reported an error: %s",
		     buf);
    }
    else {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_clamd]: %s", buf );
    }

 smtpd_clamd_cleanup_ctrl:
    apr_socket_close(sock_ctrl);

 smtpd_clamd_cleanup:
    apr_pool_destroy(pool);

    return ret;
}

static void register_hooks(apr_pool_t *p)
{
    APR_OPTIONAL_HOOK(smtpd, data_post, smtpd_clamd, NULL, NULL,
		      APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA smtpd_clamd_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                          /* dir config creater */
    NULL,                          /* dir merger --- default is to override */
    smtpd_clamd_config_new,   /* server config */
    smtpd_clamd_config_merge, /* merge server config */
    smtpd_clamd_cmds,         /* command apr_table_t */
    register_hooks                 /* register hooks */
};
